from datetime import timedelta
from functools import partial
from time import time
from typing import Optional, Dict, Any, List, Union

from flasgger import swag_from
from flask import Flask, Blueprint

from avatar_utils.healthz.checker_base import CheckerBase
from avatar_utils.healthz.utils import make_app_info
from avatar_utils.validation.schemas import RootResponseSchema

blueprint = Blueprint('healthz', __name__)

FLASGGER_INFO_DECORATOR = swag_from(dict(description='General information about service',
                                         tags=[blueprint.name],
                                         responses={200: {'description': 'Success response',
                                                          'schema': RootResponseSchema}}))
FLASGGER_PING_DECORATOR = swag_from(dict(description='Ping route',
                                         tags=[blueprint.name]))
FLASGGER_HEALTHZ_DECORATOR = swag_from(dict(description='Health checking',
                                            tags=[blueprint.name]))


def make_info_route(app_name: str,
                    app_version: Optional[str]) -> Dict[str, Any]:
    return make_app_info(app_name=app_name, app_version=app_version)


def make_healthz_route(health_checkers: List[CheckerBase]):

    response = dict()
    response.update({
        'health_checkers': list(),
        'healthy': True
    })

    for health_checker in health_checkers:
        started_at = time()
        passed = health_checker.call()
        finished_at = time() - started_at
        elapsed_time = timedelta(seconds=finished_at)

        if not passed:
            response['all_done'] = False

        response['health_checkers'].append({
            'name': health_checker.name,
            'passed': passed,
            'elapsed': str(elapsed_time),
        })

    return response


def make_healthz_blueprint(app_name: str,
                           app_version: Optional[str] = None,
                           flask_app: Optional[Flask] = None,
                           url_prefix: Optional[str] = None,
                           health_checkers: Optional[Union[CheckerBase,
                                                           List[CheckerBase]]] = None) -> Blueprint:

    if not health_checkers:
        health_checkers = list()

    if isinstance(health_checkers, CheckerBase):
        health_checkers = [health_checkers]

    info_route = partial(make_info_route, app_name=app_name, app_version=app_version)
    healthz_route = partial(make_healthz_route, health_checkers=health_checkers)

    blueprint.add_url_rule('/', 'info_route', FLASGGER_INFO_DECORATOR(info_route))
    blueprint.add_url_rule('/ping', 'ping_route', FLASGGER_PING_DECORATOR(lambda: 'pong'))
    blueprint.add_url_rule('/healthz', 'healthz_route', FLASGGER_HEALTHZ_DECORATOR(healthz_route))

    if flask_app is not None:
        flask_app.register_blueprint(blueprint,
                                     url_prefix=url_prefix)

    return blueprint



from logging import getLogger

from flask_sqlalchemy import SQLAlchemy

from avatar_utils.healthz.checker_base import CheckerBase

logger = getLogger(__name__)


class PostgresqlChecker(CheckerBase):

    def __init__(self, db: SQLAlchemy):
        super(PostgresqlChecker, self).__init__(name='postgresql', db=db)

    def checker(self, db: SQLAlchemy) -> bool:
        is_passed: bool = False

        try:
            a = db.engine.execute('SELECT 1')
            is_passed = True
        except Exception as err:  # noqa
            logger.error('Health check failed:\n{err}')

        return is_passed



from json import loads, dumps
from typing import Union, Any

from app import db


class JsonCastMixin:

    def __init__(self):
        try:
            self.db = db
        except:  # noqa
            self.db = None

    def str_field_json_loads(
            self,
            field_name: str,
            **kwargs,
    ) -> Union[str, list, dict]:

        value: str = getattr(self, field_name)
        if not isinstance(value, str):
            raise KeyError

        return loads(value, **kwargs)

    def filter_by_json_str_field(
            self,
            field_name: str,
            value: Union[str, list, tuple, dict],
            **kwargs: Any,
    ) -> object:

        stored_value: str = getattr(self, field_name)
        if not isinstance(stored_value, str):
            raise KeyError
        kwargs.update({field_name: dumps(value)})

        return db.query.filter_by(**kwargs)



from logging import getLogger

logger = getLogger(__name__)


class Env:
    protocol: str
    host: str
    port: int

    def __init__(self, protocol: str = 'http://', host: str = 'localhost', port: int = 5000):
        self.protocol = protocol
        self.host = host
        self.port = port

        logger.debug(f'{protocol}, {host}, {port}')

    @property
    def base_url(self):
        url = f'{self.protocol}{self.host}'
        if self.port:
            url = f'{url}:{self.port}'

        return url


class DevEnv(Env):

    def __init__(self, protocol: str = 'http://', host: str = 'localhost', port: int = 8000):
        logger.debug(f'dev: {protocol}, {host}, {port}')
        super().__init__(protocol, host, port)


class TestEnv(Env):

    def __init__(self, protocol: str = 'https://', host: str = 'localhost', port: int = None):
        super().__init__(protocol, host, port)


class ProdEnv(Env):

    def __init__(self, protocol: str = 'https://', host: str = 'localhost', port: int = None):
        super().__init__(protocol, host, port)



from dataclasses import dataclass as python_dataclass, field
from typing import Any, List

from marshmallow import ValidationError
from marshmallow_dataclass import dataclass

from avatar_utils.objects.abstracts.serializable import Serializable


@dataclass
@python_dataclass
class AbstractQuestionWithOptions(Serializable):

    def retrieve_answers(self):
        answers = []
        options_len = len(self.options)
        for answer_id in self.answers:
            if answer_id >= options_len:
                raise ValidationError({
                    'answers': f'No existing option has been picked! Answer id: {answer_id}. Options length: {options_len}.'
                })

            answer = self.options[answer_id]
            answers.append(answer)

        return answers

@dataclass
@python_dataclass
class ChoiseQuestion(AbstractQuestionWithOptions):
    options: List[Any] = field(default_factory=list)
    answers: List[Any] = field(default_factory=list)
    label: str = None

    repr_type: str = None

    id: int = None


@dataclass
@python_dataclass
class SelectQuestion(AbstractQuestionWithOptions):
    options: List[Any] = field(default_factory=list)
    answers: List[Any] = field(default_factory=list)
    label: str = None

    repr_type: str = None

    id: int = None


@dataclass
@python_dataclass
class MultipleSelectQuestion(AbstractQuestionWithOptions):
    options: List[Any] = field(default_factory=list)
    answers: List[Any] = field(default_factory=list)
    label: str = None

    repr_type: str = None

    id: int = None


@dataclass
@python_dataclass
class CheckBoxQuestion(AbstractQuestionWithOptions):
    options: List[Any] = field(default_factory=list)
    answers: List[Any] = field(default_factory=list)
    label: str = None

    repr_type: str = None

    id: int = None


@dataclass
@python_dataclass
class RangeQuestion(Serializable):
    answers: List[Any] = field(default_factory=list)
    min_value: float = None
    max_value: float = None
    label: str = None

    repr_type: str = None

    id: int = None


@dataclass
@python_dataclass
class TextFieldQuestion(Serializable):
    answers: List[Any] = field(default_factory=list)
    shadow_title: str = None
    validation_type: str = None
    label: str = None

    repr_type: str = None

    id: int = None



import logging
from unittest import TestCase

from avatar_utils.tests.AbstractApp import AbstractApp

logger = logging.getLogger(__name__)


class UnitTest(TestCase, AbstractApp):
    app = None
    db = None

    def setUp(self):
        TestCase.setUp(self)
        try:
            print('SQLALCHEMY_DATABASE_URI: %s' % self.app.config['SQLALCHEMY_DATABASE_URI'])
        except KeyError:
            pass

        self.assertTrue(self.app)
        self.assertTrue(self.client)
        self.assertTrue(self.app.config['TESTING'])

        self.app_context = self.app.app_context()
        self.app_context.push()

        if self.db:
            self.db.create_all()

        super().setUp()

    def tearDown(self):
        if self.db:
            self.db.session.remove()
            self.db.drop_all()

        self.app_context.pop()

        TestCase.tearDown(self)



import logging
from json import dumps as json_dumps

import flask

JSON_DECODE_ATTRS = dict(indent=2, ensure_ascii=False)

logger = logging.getLogger(__name__)


class AbstractApp:
    client = None
    headers = {'Content-type': 'application/json'}

    def post(self, url, data=dict()) -> flask.Response:
        return self.client.post(url, json=data, headers=self.headers)

    def get(self, url, data=dict()) -> flask.Response:
        return self.client.get(url, json=data, headers=self.headers)

    def register(self, username, email, password, confirm) -> flask.Response:
        return self.post('/register', data=dict(username=username, email=email, password=password, confirm=confirm))

    def login(self, username, password) -> flask.Response:
        return self.post('/login', data=dict(username=username, password=password))

    def logout(self) -> flask.Response:
        return self.post('/logout')

    def delete(self, user_id, password):
        return self.post('/user/delete', data=dict(user_id=user_id, password=password))

    def rprint(self, response):
        if not isinstance(response, flask.Response):
            logger.warning(f'rprint prints only flask.Response class data')
            return

        json = response.json
        if json:
            logger.debug(json)
        else:
            logger.debug(response.data.decode())

    def json_dumps(self, data):
        return json_dumps(data, **JSON_DECODE_ATTRS)



from typing import Optional, Tuple, Dict

from flask import request
from requests import Session

from avatar_utils.sso_helper.constants import (
    AUTH_HEADER_NAME,
    DEFAULT_TOKEN_TYPE,
)


class FlaskAuthHeader:

    @staticmethod
    def from_request() -> Optional[str]:

        return request.headers.get(AUTH_HEADER_NAME)

    @classmethod
    def request_token(cls) -> Optional[str]:

        value, _ = cls.extract()
        return value

    @classmethod
    def request_token_type(cls) -> Optional[str]:

        _, value = cls.extract()
        return value

    @classmethod
    def extract(cls) -> Tuple[Optional[str], Optional[str]]:

        raw_value = cls.from_request()
        token_data = raw_value.split(' ') if raw_value else (None, None)
        token_type, token_string = token_data
        return token_string, token_type

    @classmethod
    def make_header_value(cls,
                          token_string: Optional[str] = None,
                          token_type: Optional[str] = None) -> Optional[str]:

        if not token_type and token_string and token_string.startswith(DEFAULT_TOKEN_TYPE):
            token_type, token_string = token_string.split(' ')

        if not token_type and token_string:
            token_type = DEFAULT_TOKEN_TYPE

        if not token_type and not token_string:
            token_string, token_type = cls.extract()

        return f'{token_type} {token_string}'

    @classmethod
    def make_header_dict(cls,
                         token_string: Optional[str] = None,
                         token_type: Optional[str] = None) -> Dict[str, Optional[str]]:

        value = cls.make_header_value(token_string, token_type)
        return {AUTH_HEADER_NAME: value}

    @classmethod
    def make_session(cls,
                     token_string: Optional[str] = None,
                     token_type: Optional[str] = None) -> Optional[Session]:
        session = Session()
        header = cls.make_header_dict(token_string=token_string, token_type=token_type)
        if header[AUTH_HEADER_NAME] is not None:
            session.headers.update(header)
            return session
